/*
 * If CONTENT_LENGTH is set:
 *	Process a form passed using the POST method, extracting
 *	the "callsign" and "message" fields and adding them to the
 *	end of the file MESSAGE, with proper HTML formatting.
 *
 * Always:
 *	Output SKINFILE to stdout, replacing any occurence of ''
 *	with the contents of the file CHATFILE, and any occurence
 *	of '' with the last callsign, if present.
 *
 * Copyright (c) David Olofson, 2002.
 * This file is released under the terms of the GNU GPL.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <fcntl.h>

#include "chat.h"


/*---------------------------------------------------------------------
	Process and show the chat skin
---------------------------------------------------------------------*/

static void print_lines(void)
{
	int f;
	char *buf;
	struct stat st;

	f = open(CHATFILE, O_RDONLY);
	if(!f)
		exit(1);
	if(flock(f, LOCK_EX) < 0)
		exit(1);

	if(fstat(f, &st) < 0)
		exit(1);

	buf = malloc(st.st_size);
	if(!buf)
		exit(1);

	if(read(f, buf, st.st_size) != st.st_size)
		puts("<font color=\"#ff9966\">Chat file read error!<br>");
	else
		fwrite(buf, st.st_size, 1, stdout);

	flock(f, LOCK_UN);
	close(f);
}


static int print_page(const char *callsign)
{
	char c;
	FILE *f;

	puts("Content-type: text/html\n");

	f = fopen(SKINFILE, "rb");
	if(!f)
		return 1;

	while((c = fgetc(f)) != EOF)
		switch(c)
		{
		  case '':
			print_lines();
			break;
		  case '':
			fputs(callsign, stdout);
			break;
		  default:
			putchar(c);
		}
	fclose(f);
	return 0;
}



/*---------------------------------------------------------------------
	Generic function to pass any HTML file to the output
---------------------------------------------------------------------*/

static void print_file(const char *name)
{
	char c;
	FILE *f;

	puts("Content-type: text/html\n");

	f = fopen(name, "rb");
	if(!f)
		exit(1);
	while((c = fgetc(f)) != EOF)
		putchar(c);
	fclose(f);
}


/*---------------------------------------------------------------------
	Extract data from form POST submissions
---------------------------------------------------------------------*/

static inline int hex2int(const char *s)
{
	int i, v = 0;
	for(i = 0; i < 2; ++i)
	{
		v <<= 4;
		if((s[i] >= '0') && (s[i] <= '9'))
			v += s[i] - '0';
		else if((s[i] >= 'A') && (s[i] <= 'F'))
			v += s[i] - 'A' + 10;
		else if((s[i] >= 'a') && (s[i] <= 'f'))
			v += s[i] - 'a' + 10;
	}
	return v;
}


static char *getfield(const char *buf, const char *name)
{
	int len, stop;
	char *cpi, *cpo, *r;

	/* Find field by name */
	cpi = NULL;
	while(!cpi)
	{
		cpi = strstr(buf, name);
		if(!cpi)
			return NULL;

		if('=' == cpi[strlen(name)])
		{
			cpi += strlen(name);
			++cpi;
			break;
		}

		buf = cpi + strlen(name);
		cpi = NULL;
	}

	/* Allocate buffer of sufficient size */
	cpo = cpi;
	len = 0;
	while(('&' != *cpo) && *cpo)
		++len, ++cpo;
	r = malloc(len + 1);
	if(!r)
		exit(1);

	/* Decode data */
	cpo = r;
	stop = 0;
	while(1)
	{
		switch(*cpi)
		{
		  case '+':
			*cpo = ' ';
			break;
		  case '%':
			++cpi;
			*cpo = hex2int(cpi);
			++cpi;
			break;
		  case 0:
		  case '&':
			stop = 1;
			break;
		  default:
			*cpo = *cpi;
		}
		if(stop)
			break;
		++cpi;
		++cpo;
	}
	*cpo = 0;

	return r;
}



/*---------------------------------------------------------------------
	Add a message to the chatfile
---------------------------------------------------------------------*/

static void add_message(const char *callsign, const char *message)
{
	int msg = open(CHATFILE, O_CREAT | O_APPEND | O_WRONLY);
	if(!msg)
		exit(1);
	if(flock(msg, LOCK_EX) < 0)
		exit(1);
	write(msg, "<b>", 3);
	write(msg, callsign, strlen(callsign));
	write(msg, ":</b> ", 6);
	write(msg, message, strlen(message));
	write(msg, "<br>", 4);
	flock(msg, LOCK_UN);
	close(msg);
}


/*---------------------------------------------------------------------
	Main chat program
---------------------------------------------------------------------*/

int main(int argc, char * argv[])
{
	int len;
	char *s, *callsign, *message;

	s = getenv("CONTENT_LENGTH");
	if(!s)
		return print_page("Nobody");

	len = atoi(s);
	s = malloc(len+1);
	if(!s)
		exit(1);

	if(fgets(s, len+1, stdin) < 0)
		exit(1);

	callsign = getfield(s, "callsign");
	message = getfield(s, "message");
	free(s);

	if(!callsign || !*callsign)
	{
		print_file(CALLSIGN_ERROR);
		return 0;
	}

	if(message && *message)
		add_message(callsign, message);

	return print_page(callsign);
}
